comment		^

	LZW Compression and Decompression using 16 bit codes (dictionary = 0xFFFF entries)
	Dictionary entries are max 2 codes (entry = dword)

		^

cDictionary	=	0FFFFh	;number of entries in Dictionary
.data
	FirstChar	dd	?
.code
	Decode	proto

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	LZW_Compress	proc, Source: dword, Destination: dword, cbLength: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	
	local pDictionary: 	dword
	
	;--- Create Dictionary space ---
	invoke VirtualAlloc, NULL, cDictionary*4, MEM_COMMIT, PAGE_READWRITE
	
	;--- Init ---
	mov edx, eax		;edx = pDictionary
	mov esi, Source		;esi = ptr source file
	add cbLength, esi	;ptr EOF
	mov edi, Destination	;edi = ptr output
	mov ecx, 100h		;ecx = index of next available Dictionary code
	mov ebx, ecx
	shl ebx, 2
	add ebx, edx		
	sub ebx, 4		;ebx = ptr last entry
	xor eax, eax		;eax = wk
	cld
		
	;--- Load (k) with first Character ---
	lodsb

CompressionLoop:
	;--- w<<k ---
	shl eax, 16
	;--- get (k) ---
	lodsb		;eax = wk
	;--- Search Dictionary for wk ---
	push edi
	push ecx
	mov edi, ebx	;edi = ptr last entry
	std
	repne scasd
	cld		
	.if ecx	> 0FFh	;match found, make k = code
		mov eax, ecx		;k = wk
		pop ecx
		pop edi
	
	.else		;not found
		;--- Add to Dictionary ---
		pop ecx
		.if ecx != cDictionary	;room left for new entry
			mov [edx+ecx*4], eax	;save wk in Dictionary
			add ebx, 4		;ptr last entry
			inc ecx			;advance Dictionary index
			
		.endif

		;--- Output w ---
		pop edi
		push eax
		shr eax, 16
		stosw
		pop eax

	.endif
	
		cmp esi, cbLength
		jne CompressionLoop
		stosw			;send final k
		sub edi, Destination
		push edi
		invoke VirtualFree, edx, NULL, MEM_RELEASE
		pop eax		;return byte count of Compressed file
		ret

LZW_Compress endp


;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	LZW_DeCompress	proc, Source: dword, Destination: dword, cbLength: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	
	local	DictionaryUpdated:	dword

	;--- Create Dictionary space ---
	invoke VirtualAlloc, NULL, cDictionary*4, MEM_COMMIT, PAGE_READWRITE
	

	;--- Init ---
	mov edx, eax		;edx = ptr Dictionary
	mov esi, Source		;esi = ptr source file
	add cbLength, esi	;ptr EOF
	mov edi, Destination	;edi = ptr output
	mov ecx, 100h		;ecx = index of next available Dictionary code
	xor eax, eax		;eax = wk
	cld

	;--- Load (k) with first Character ---
	lodsw
	stosb

DeCompressLoop:

	mov DictionaryUpdated, FALSE

	;--- w<<k ---
	shl eax, 16
	
	;--- Read k ---
	lodsw

	;--- If k = next uncreated Dictionary entry ---
	.if ax == cx
		push ax			;k
		mov ebx, FirstChar	
		mov ax, bx		;eax = new entry
		mov [edx+ecx*4], eax	;add new entry to Dictionary
		pop ax
		inc ecx			;update Dictionary Index
		mov DictionaryUpdated, TRUE
	.endif

	;--- If k is less than 256 then output k ---
	.if ax < 100h
		stosb
		push eax
		and eax, 0FFh
		mov FirstChar, eax
		pop eax

	;--- If k is greater than 255, its a code. Output Dictionary entry for k 
	.else
		push eax
		and eax, 0FFFFh		;eax = code
		mov eax, [edx + eax*4]	;eax = entry 
		mov FirstChar, -1	;invalidate
		mov ebx, 0		;flag for Decode
		invoke Decode		;output entry
		pop eax
	.endif

	;--- Add w + (FirstChar of entry for k) to Dictionary ---
	.if (DictionaryUpdated == FALSE) && (ecx != cDictionary)
		push eax
		mov ebx, FirstChar
		movzx ax, bl
		mov [edx + ecx*4], eax
		inc ecx
		pop eax
	.endif
	
	cmp esi, cbLength
	jne DeCompressLoop

	;--- Return byte length of uncompressed file ---
	sub edi, Destination
	mov eax, edi
	ret

LZW_DeCompress endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	Decode	proc	
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	;ebx = 0 on first pass
	;ebx = NOT 0 on second pass
	
	ror eax, 16	;ax = w on first pass, k on second
	.if ax < 100h	;w = char
		.if FirstChar == -1
			push eax
			and eax, 0FFh
			mov FirstChar, eax
			pop eax
		.endif
		stosb
	.else		;w = code
		push eax	;save current wk
		and eax, 0FFFFh
		mov eax, [edx + eax*4]	;eax = Dictionary entry for code
		invoke Decode		;output entry
		pop eax	
	.endif

	not ebx
	.if ebx
		invoke Decode
	.endif
		ret

Decode endp












	